/*
 * Copyright (C) 2003 Sebastian Krahmer.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Sebastian Krahmer.
 * 4. The name Sebastian Krahmer may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <cstring>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <arpa/inet.h>

#define __FAVOR_BSD
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip6.h>
#include <cerrno>
#include <iostream>
#include <unistd.h>
#include <cstdio>
#include "packet.h"
#include "misc.h"
#include "strategy.h"


namespace loaded {

using namespace std;

//Virtual connections src -> dst
map<sockaddr_in, sockaddr_in, sockaddrLess> VConn4;
map<sockaddr_in6, sockaddr_in6, less<sockaddr_in6> > VConn6;



sockaddr_in VIP4;
sockaddr_in6 VIP6;


int fix_ipv4(unsigned char *packet, size_t p_len, const sockaddr *saddr)
{
	iphdr *iph = NULL;
	tcphdr *tcph = NULL;
	udphdr *udph = NULL;

	struct {
		u_int32_t saddr, daddr;
		u_int8_t zero, proto;
		u_int16_t len;
	} pseudo;
	unsigned char *calc_here = NULL;

	iph = (iphdr *)packet;
	
	if (iph->daddr == VIP4.sin_addr.s_addr)
		iph->daddr = ((sockaddr_in*)saddr)->sin_addr.s_addr;
	else {
		iph->saddr = VIP4.sin_addr.s_addr;
	}

	// some nose is sending evil packets
	if (p_len - (iph->ihl<<2) < 0)
		return 0;

	switch (iph->protocol) {
	case IPPROTO_TCP:
		pseudo.zero = 0;
		pseudo.saddr = iph->saddr;
		pseudo.daddr = iph->daddr;
		pseudo.proto = iph->protocol;
		pseudo.len = htons(p_len-(iph->ihl<<2));
		tcph = (struct tcphdr*)(packet + (iph->ihl<<2));

		calc_here = new unsigned char [sizeof(pseudo)+p_len];
		tcph->th_sum = 0;
		memcpy(calc_here, &pseudo, sizeof(pseudo));
		memcpy(calc_here+sizeof(pseudo), tcph, p_len - (iph->ihl<<2));

		tcph->th_sum = in_cksum((unsigned short*)calc_here,
		                        p_len + sizeof(pseudo) - (iph->ihl<<2));

		delete [] calc_here;
		break;
	case IPPROTO_UDP:
		pseudo.zero = 0;
		pseudo.saddr = iph->saddr;
		pseudo.daddr = iph->daddr;
		pseudo.proto = iph->protocol;
		pseudo.len = htons(p_len-(iph->ihl<<2));
		udph = (struct udphdr*)(packet + (iph->ihl<<2));

		calc_here = new unsigned char [sizeof(pseudo)+p_len];
		udph->uh_sum = 0;
		memcpy(calc_here, &pseudo, sizeof(pseudo));
		memcpy(calc_here+sizeof(pseudo), udph, p_len - (iph->ihl<<2));

		udph->uh_sum = in_cksum((unsigned short*)calc_here,
		                        p_len + sizeof(pseudo) - (iph->ihl<<2));

		delete [] calc_here;
		break;
	case IPPROTO_ICMP:

		// No checksum recalculation needed
		break;
	default:
		return -1;
	}

	iph->check = 0;
	iph->check = in_cksum((unsigned short*)iph, iph->ihl<<2);

	return 0;
}


int fix_ipv6(unsigned char *packet, size_t plen, const sockaddr *saddr)
{

	ip6_hdr *ip6h = (ip6_hdr*)packet;

	switch (ip6h->ip6_nxt) {
	case IPPROTO_TCP:
		break;
	case IPPROTO_ICMPV6:
		break;
	default:
		return -1;
	}


	return 0;
}


int Packet::init()
{
	if ((h = ipq_create_handle(0, PF_INET)) == NULL) {
		err = "Packet::init::ipq_create_handle:";
		err += strerror(errno);
		return -1;
	}
	if (ipq_set_mode(h, IPQ_COPY_PACKET, MTU) < 0) {
		err = "Packet::init::ipq_set_mode";
		err += strerror(errno);
		return -1;
	}

	if (strategy == "weighted")
		choose_ip4 = choose_ip4_weighted;
	else if (strategy == "rr")
		choose_ip4 = choose_ip4_rr;

	return 0;
}


int Packet::recv()
{

	if ((ipq_read(h, packet, MTU, 0)) < 0) {
		err = "Packet::recv::ipq_read";
		err += ipq_get_msgerr(packet);
		if (errno)
			err += strerror(errno);
		return -1;
	}

	switch (ipq_message_type(packet)) {
	case NLMSG_ERROR:
		err = "Received error message ";
		err += ipq_get_msgerr(packet);
		return -1;
	case IPQM_PACKET:
		pm = ipq_get_packet(packet);
		break;
	default:
		err = "Unknown error.";
		return -1;
	}

	return 0;
}


int Packet::send(const sockaddr *saddr, int af)
{
	sockaddr_in null_addr;
	memset(&null_addr, 0, sizeof(null_addr));

	if (af == AF_INET) {
		fix_ipv4(pm->payload, pm->data_len, saddr);

		sockaddr_in sin;
		memset(&sin, 0, sizeof(sin));
		sin.sin_addr = ((sockaddr_in*)saddr)->sin_addr;

		// don't count packets which were not destinated for
		// the virtual IP (i.e. packets coming back from
		// real servers) in which case Packet::balance() returned
		// 0 in b4 which was then passed to Packet::send()
		if (memcmp(&null_addr, &sin, sizeof(sin)) != 0)
			++packetCount4[sin];
	} else if (af == AF_INET6) {
		fix_ipv6(pm->payload, pm->data_len, saddr);
	} else {
		err = "Packet::send: Unknown protocol (neither IPv4 nor IPv6).";
		return -1;
	}

	if (ipq_set_verdict(h, pm->packet_id, NF_ACCEPT, pm->data_len,
	                    pm->payload) < 0) {
		err = "Packet::send::ipq_set_verdict";
		err += strerror(errno);
		return -1;
	}

	return 0;
}


int Packet::balance(sockaddr_in &sin, sockaddr_in6 &sin6, int af)
{
	sockaddr_in null_addr;
	memset(&null_addr, 0, sizeof(null_addr));

	if (af == AF_INET) {
		sockaddr_in s;
		iphdr *iph = (iphdr*)pm->payload;

		// We dont need to balance if it isnt
		// for the virtual IP anyways
		// Packet::send()/fix_chksums() will handle this correctly
		if (iph->daddr != VIP4.sin_addr.s_addr) {
//			in_addr ia = { iph->daddr };
//			cerr<<"A"<<inet_ntoa(ia)<<endl;
			sin = null_addr;
			return 0;
		}

		memset(&s, 0, sizeof(s));
		s.sin_addr.s_addr = iph->saddr;

/*
		map<sockaddr_in, sockaddr_in>::iterator j = VConn4.begin();
		for (; j != VConn4.end(); ++j) {
			cerr<<inet_ntoa(j->first.sin_addr)<<"->"<<inet_ntoa(j->second.sin_addr)<<endl;
		}
*/

		// if we already chose an IP, use it
		if (VConn4.find(s) != VConn4.end()) {
			sin = VConn4[s];
			return 0;
		}


		sin = choose_ip4();

		VConn4[s] = sin;
		return 0;
	} else if (af == AF_INET6) {
		cerr<<"AAAA\n";
	} else {
		cerr<<"B\n";
	}
	return 0;
}

}

